from dataclasses import dataclass
from typing import Hashable

import igraph as ig

from nasap_net.exceptions import NasapNetError
from ..exceptions import NoIsomorphismFoundError


@dataclass(frozen=True)
class Colors:
    v_color1: tuple[int, ...]
    v_color2: tuple[int, ...]
    e_color1: tuple[int, ...]
    e_color2: tuple[int, ...]


def color_vertices_and_edges(g1: ig.Graph, g2: ig.Graph) -> Colors:
    try:
        v_color1, v_color2 = _vertex_color_lists(g1, g2)
    except _NotIsomorphicError:
        raise NoIsomorphismFoundError() from None

    try:
        e_color1, e_color2 = _edge_color_lists(g1, g2)
    except _NotIsomorphicError:
        raise NoIsomorphismFoundError() from None

    return Colors(
        v_color1=tuple(v_color1),
        v_color2=tuple(v_color2),
        e_color1=tuple(e_color1),
        e_color2=tuple(e_color2),
    )


class _NotIsomorphicError(NasapNetError):
    pass


def _vertex_color_lists(
        g1: ig.Graph, g2: ig.Graph) -> tuple[list[int], list[int]]:
    def _readable_v_color(vertex: ig.Vertex) -> Hashable:
        return vertex['core_or_site'], vertex['comp_kind']

    colors1 = {_readable_v_color(v) for v in g1.vs}
    colors2 = {_readable_v_color(v) for v in g2.vs}
    if colors1 != colors2:
        raise _NotIsomorphicError()

    color_to_int: dict[Hashable, int] = {c: i for i, c in enumerate(colors1)}
    color_list1 = [color_to_int[_readable_v_color(v)] for v in g1.vs]
    color_list2 = [color_to_int[_readable_v_color(v)] for v in g2.vs]
    return color_list1, color_list2


def _edge_color_lists(
        g1: ig.Graph, g2: ig.Graph) -> tuple[list[int], list[int]]:
    def _readable_edge_color(edge: ig.Edge) -> Hashable | None:
        return edge['aux_kind'] if 'aux_kind' in edge.attributes() else None

    colors1 = {_readable_edge_color(e) for e in g1.es}
    colors2 = {_readable_edge_color(e) for e in g2.es}
    if colors1 != colors2:
        raise _NotIsomorphicError()

    color_to_int: dict[Hashable | None, int] = {
        c: i for i, c in enumerate(colors1)}
    color_list1 = [color_to_int[_readable_edge_color(e)] for e in g1.es]
    color_list2 = [color_to_int[_readable_edge_color(e)] for e in g2.es]
    return color_list1, color_list2



from .exceptions import NoIsomorphismFoundError
from .get_isomorphism import get_all_isomorphisms, get_isomorphism
from .is_isomorphic import is_isomorphic
from .isomorphic_assembly_search import AssemblyNotFoundError, \
    IsomorphicAssemblyFinder
from .models import Isomorphism
from .unique import extract_unique_assemblies_by_isomorphism



import pytest

from nasap_net.models import Assembly, AuxEdge, Bond, Component
from nasap_net.yaml import dump, dump_to_file


@pytest.fixture
def sample_assemblies():
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    M_aux = Component(
        kind='M(aux)', sites=[0, 1, 2],
        aux_edges=[AuxEdge(0, 1), AuxEdge(0, 2, kind='cis')]
    )

    assemblies = [
        Assembly(components={'X0': X}, bonds=[]),
        Assembly(
            id_='MX2',
            components={'X0': X, 'M0': M, 'X1': X},
            bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]
        ),
        Assembly(
            components={'M0': M_aux, 'X0': X, 'X1': X, 'X2': X},
            bonds=[
                Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0), Bond('M0', 2, 'X2', 0)
            ]
        ),
    ]

    return assemblies


def test_writes_file_and_contents(tmp_path, sample_assemblies):
    assemblies = sample_assemblies
    out = tmp_path / "out.yaml"

    # write
    dump_to_file(assemblies, out, overwrite=False, verbose=False)

    # file exists and contents match dump()
    assert out.exists()
    text = out.read_text()
    assert text == dump(assemblies)


def test_raises_if_exists_and_no_overwrite(tmp_path, sample_assemblies):
    assemblies = sample_assemblies
    out = tmp_path / "already.yaml"
    out.write_text("old")

    try:
        dump_to_file(assemblies, out, overwrite=False, verbose=False)
        raised = False
    except FileExistsError:
        raised = True

    assert raised is True
    # original content unchanged
    assert out.read_text() == "old"


def test_overwrite_ok(tmp_path, sample_assemblies):
    assemblies = sample_assemblies
    out = tmp_path / "to_overwrite.yaml"
    out.write_text("old content")

    dump_to_file(assemblies, out, overwrite=True, verbose=False)

    assert out.read_text() == dump(assemblies)


def test_creates_parent_dirs(tmp_path, sample_assemblies):
    assemblies = sample_assemblies
    out = tmp_path / "nested" / "a" / "dump.yaml"

    dump_to_file(assemblies, out, overwrite=False, verbose=False)

    assert out.exists()
    assert (tmp_path / "nested" / "a").exists()


def test_verbose_prints_saved_message(tmp_path, capsys, sample_assemblies):
    assemblies = sample_assemblies
    out = tmp_path / "v.yaml"

    dump_to_file(assemblies, out, overwrite=False, verbose=True)

    captured = capsys.readouterr()
    # message should include path
    assert 'Saved! -->' in captured.out
    assert str(out) in captured.out



import pytest

from nasap_net.models import Assembly, AuxEdge, Bond, Component
from nasap_net.yaml import dump, load_from_file


@pytest.fixture
def sample_assemblies():
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    M_aux = Component(
        kind='M(aux)', sites=[0, 1, 2],
        aux_edges=[AuxEdge(0, 1), AuxEdge(0, 2, kind='cis')]
    )

    assemblies = [
        Assembly(components={'X0': X}, bonds=[]),
        Assembly(
            id_='MX2',
            components={'X0': X, 'M0': M, 'X1': X},
            bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]
        ),
        Assembly(
            components={'M0': M_aux, 'X0': X, 'X1': X, 'X2': X},
            bonds=[
                Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0), Bond('M0', 2, 'X2', 0)
            ]
        ),
    ]

    return assemblies


def test_reads_file_and_contents(tmp_path, sample_assemblies):
    assemblies = sample_assemblies
    f = tmp_path / "in.yaml"
    # write YAML using dump()
    f.write_text(dump(assemblies))

    loaded = load_from_file(f, strict=True, verbose=False)

    assert isinstance(loaded, list)
    assert loaded == assemblies


def test_raises_if_missing_and_strict(tmp_path):
    missing = tmp_path / "noexist.yaml"
    with pytest.raises(FileNotFoundError):
        load_from_file(missing, strict=True, verbose=False)


def test_returns_empty_if_missing_and_not_strict(tmp_path):
    missing = tmp_path / "noexist.yaml"
    res = load_from_file(missing, strict=False, verbose=False)
    assert res == []


def test_verbose_prints_loaded_message(tmp_path, capsys, sample_assemblies):
    assemblies = sample_assemblies
    f = tmp_path / "v_in.yaml"
    f.write_text(dump(assemblies))

    load_from_file(f, strict=True, verbose=True)

    captured = capsys.readouterr()
    assert 'Loaded' in captured.out
    assert str(f) in captured.out



from collections.abc import Iterable, Mapping
from typing import Any

from nasap_net.types import ID
from .sub_fragment import create_sub_fragment
from .unique import extract_unique_fragments
from ..models import Fragment


def get_unique_starting_fragments(
        template: Fragment,
        symmetry_operations: Iterable[Mapping[Any, ID]] | None = None
) -> set[Fragment]:
    raw_starting_frags = _get_raw_starting_fragments(template)
    return set(
        extract_unique_fragments(
            sorted(raw_starting_frags), symmetry_operations
        )
    )


def _get_raw_starting_fragments(template: Fragment) -> set[Fragment]:
    return {
        create_sub_fragment(template, [comp_id])
        for comp_id in sorted(template.components)
    }



from collections.abc import Iterable

from nasap_net.models import Assembly, BindingSite, Bond, Component
from nasap_net.types import ID


def cap_assemblies_with_ligand(
        fragments: Iterable[Assembly],
        component: Component,
        component_site_id: ID | None,
        metal_kinds: Iterable[str],
) -> set[Assembly]:
    """Cap all free metal sites in the assemblies with the given component"""
    if component_site_id is None:
        component_site_id = next(iter(sorted(component.site_ids)))

    return {
        cap_assembly(
            assembly=fragment,
            component=component,
            component_site_id=component_site_id,
            metal_kinds=metal_kinds,
        )
        for fragment in fragments
    }


def cap_assembly(
        assembly: Assembly,
        component: Component,
        component_site_id: ID,
        metal_kinds: Iterable[str],
) -> Assembly:
    """Add the leaving ligand to all free coordination sites of the metals"""
    free_metal_sites = get_free_metal_sites(assembly, metal_kinds)
    for i, site in enumerate(free_metal_sites):
        comp_id = f'{component.kind}{i}'
        assembly = add_component(
            assembly=assembly,
            assembly_site=site,
            component=component,
            component_site=BindingSite(comp_id, component_site_id),
            adding_component_id=comp_id,
        )
    return assembly


def get_free_metal_sites(
        assembly: Assembly,
        metal_kinds: Iterable[str]
) -> set[BindingSite]:
    free_sites: set[BindingSite] = set()
    for metal_kind in metal_kinds:
        free_sites.update(
            assembly.find_sites(
                has_bond=False,
                component_kind=metal_kind,
            )
        )
    return free_sites


def add_component(
        assembly: Assembly,
        assembly_site: BindingSite,
        component: Component,
        component_site: BindingSite,
        adding_component_id: ID,
) -> Assembly:
    components = dict(assembly.components)
    components[adding_component_id] = component
    bonds = set(assembly.bonds)
    bonds.add(Bond.from_sites(assembly_site, component_site))
    return Assembly(
        components=components,
        bonds=bonds,
        id_=assembly.id_or_none
    )



from dataclasses import dataclass
from typing import Mapping

from frozendict import frozendict

from nasap_net.models import BindingSite
from nasap_net.types import ID


@dataclass(frozen=True, init=False)
class Isomorphism:
    comp_id_mapping: frozendict[ID, ID]
    binding_site_mapping: frozendict[BindingSite, BindingSite]

    def __init__(
            self,
            comp_id_mapping: Mapping[ID, ID],
            binding_site_mapping: Mapping[BindingSite, BindingSite]
    ) -> None:
        object.__setattr__(
            self, 'comp_id_mapping', frozendict(comp_id_mapping))
        object.__setattr__(
            self, 'binding_site_mapping', frozendict(binding_site_mapping))



from collections.abc import Iterator

from ..models import Fragment, GrowingStep


def enumerate_one_step_grown_fragments(
        fragment: Fragment
) -> Iterator[Fragment]:
    for growing_step in fragment.enumerate_possible_growing_steps():
        yield grow_fragment(fragment, growing_step)


def grow_fragment(
        fragment: Fragment, step: GrowingStep
) -> Fragment:
    if step.component_to_add is None:
        new_components = fragment.components
    else:
        new_components = fragment.components | {step.component_to_add}
    new_bonds = fragment.bonds | {step.bond_to_add}
    return fragment.copy_with(components=new_components, bonds=new_bonds)



from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from functools import cached_property, total_ordering
from types import MappingProxyType
from typing import Any, Self

from frozendict import frozendict

from nasap_net.exceptions import IDNotSetError, NasapNetError
from nasap_net.types import ID
from nasap_net.utils import construct_repr
from .binding_site import BindingSite
from .bond import Bond
from .component import Component


@dataclass
class InvalidBondError(NasapNetError):
    bond: Bond
    detail: str | None = None

    def __str__(self) -> str:
        base_msg = f'Invalid bond: {self.bond}'
        if self.detail:
            return f'{base_msg} - {self.detail}'
        return base_msg


@dataclass
class InconsistentComponentError(NasapNetError):
    """Raised when there are inconsistent definitions for a component kind,
    i.e., the same kind name corresponds to different component structures.
    """
    component_kind: str
    component1: Component
    component2: Component

    def __str__(self) -> str:
        return (
            f'Inconsistent definitions for component kind '
            f'"{self.component_kind}": '
            f'Component 1: {self.component1}, Component 2: {self.component2}.'
        )


@dataclass
class ParallelBondError(NasapNetError):
    bond1: Bond
    bond2: Bond

    def __str__(self) -> str:
        return (
            f'Multiple bonds between the same pair of components '
            f'are not supported: '
            f'Bond 1: {self.bond1}, Bond 2: {self.bond2}.'
        )


@dataclass
class BondNotFoundError(NasapNetError):
    comp_id1: ID
    comp_id2: ID

    def __str__(self) -> str:
        return (
            f'No bond found between components '
            f'{self.comp_id1} and {self.comp_id2}.'
        )


@total_ordering
@dataclass(frozen=True, init=False)
class Assembly:
    """An assembly of components connected by bonds.

    Parameters
    ----------
    components : Mapping[C, nasap_net.models.component.Component[S]]
        A mapping from component IDs to their corresponding components.
    bonds : Iterable[Bond[C, S]]
        An iterable of bonds connecting the components.

    Raises
    ------
    - InconsistentComponentError
        If components with the same kind have different structures.
    - InvalidBondError
        - If any bond references a non-existent component or site.
        - If a component bonds to itself.
        - If a site is used more than once.
        - If the assembly is not connected.
    - ParallelBondError
        If there are multiple bonds between the same pair of components.

    Warnings
    --------
    - The assembly does not enforce connectivity; it is the user's
      responsibility to ensure that the assembly is connected as needed.

    Notes
    -----
    Currently, the assembly does not support parallel bonds (multiple bonds
    between the same pair of components), e.g., chelate complexes.
    """
    _components: frozendict[ID, Component]
    bonds: frozenset[Bond]
    _id: ID | None

    def __init__(
            self, components: Mapping[ID, Component],
            bonds: Iterable[Bond],
            *,
            id_: ID | None = None
            ):
        object.__setattr__(self, '_components', frozendict(components))
        object.__setattr__(self, 'bonds', frozenset(bonds))
        object.__setattr__(self, '_id', id_)
        self._validate()

    def __lt__(self, other):
        if not isinstance(other, Assembly):
            return NotImplemented
        # 1. number of components
        # 2. number of bonds
        # 3. component IDs (sorted)
        # 4. bonds (sorted)
        if len(self._components) != len(other._components):
            return len(self._components) < len(other._components)
        if len(self.bonds) != len(other.bonds):
            return len(self.bonds) < len(other.bonds)
        self_comp_ids = sorted(self._components.keys())
        other_comp_ids = sorted(other._components.keys())
        if self_comp_ids != other_comp_ids:
            return self_comp_ids < other_comp_ids
        self_bonds = sorted(self.bonds)
        other_bonds = sorted(other.bonds)
        return self_bonds < other_bonds

    def __repr__(self):
        fields: dict[str, Any] = {}
        if self._id is not None:
            fields['id'] = self._id
        fields['components'] = dict(sorted(self.component_id_to_kind.items()))
        fields['bonds'] = [bond.to_tuple() for bond in sorted(self.bonds)]
        return construct_repr(self.__class__, fields)

    @property
    def id(self) -> ID:
        """Return the ID of the assembly."""
        if self._id is None:
            raise IDNotSetError("Assembly ID is not set.")
        return self._id

    @property
    def id_or_none(self) -> ID | None:
        """Return the ID of the assembly, or None if not set."""
        return self._id

    @property
    def components(self) -> Mapping[ID, Component]:
        """Return the components in the assembly as an immutable mapping."""
        return MappingProxyType(self._components)

    @cached_property
    def component_id_to_kind(self) -> Mapping[ID, str]:
        """Return a mapping from component IDs to their kinds."""
        return MappingProxyType({
            comp_id: comp.kind for comp_id, comp
            in self._components.items()
        })

    def get_component_kind_of_site(self, site: BindingSite) -> str:
        """Return the component kind of the given binding site."""
        return self._components[site.component_id].kind

    def find_sites(
            self, *, has_bond: bool | None = None,
            component_kind: str | None = None
            ) -> frozenset[BindingSite]:
        """Return binding sites based on their bond status.
        """
        if has_bond is None and component_kind is None:
            return self._all_sites

        sites = set()
        for site in self._all_sites:
            if has_bond is not None:
                if has_bond != self.has_bond(site):
                    continue
            if component_kind is not None:
                if self.get_component_kind_of_site(site) != component_kind:
                    continue
            sites.add(site)
        return frozenset(sites)

    def has_bond(self, site: BindingSite) -> bool:
        """Check if a binding site has a bond."""
        return site in self._sites_with_bond

    def has_bond_between_components(self, comp_id1: ID, comp_id2: ID) -> bool:
        """Check if there is a bond between two components."""
        return frozenset({comp_id1, comp_id2}) in self._component_connection

    def get_bond_by_comp_ids(self, comp_id1: ID, comp_id2: ID) -> Bond:
        """Return the bond between two components.

        Parameters
        ----------
        comp_id1 : ID
            The ID of the first component.
        comp_id2 : ID
            The ID of the second component.

        Returns
        -------
        Bond
            The bond between the two components.

        Raises
        ------
        BondNotFoundError
            If there is no bond between the two components.

        Notes
        -----
        This method assumes that there is no parallel bonds between the
        same pair of components.
        """
        for bond in self.bonds:
            if bond.component_ids == frozenset({comp_id1, comp_id2}):
                return bond
        raise BondNotFoundError(comp_id1=comp_id1, comp_id2=comp_id2)

    def add_bond(self, site1: BindingSite, site2: BindingSite):
        """Return a new assembly with an additional bond."""
        new_bond = Bond.from_sites(site1, site2)
        new_bonds = set(self.bonds)
        new_bonds.add(new_bond)
        return self.copy_with(bonds=new_bonds)

    def remove_bond(self, site1: BindingSite, site2: BindingSite):
        """Return a new assembly with a bond removed."""
        bond_to_remove = Bond.from_sites(site1, site2)
        new_bonds = set(self.bonds)
        new_bonds.remove(bond_to_remove)
        return self.copy_with(bonds=new_bonds)

    def copy_with(
            self,
            *,
            components: Mapping[ID, Component] | None = None,
            bonds: Iterable[Bond] | None = None,
            id_: ID | None = None,
            ) -> Self:
        """Return a copy of the assembly with optional modifications."""
        if components is None:
            components = self.components
        if bonds is None:
            bonds = self.bonds
        return self.__class__(
            components=components, bonds=bonds, id_=id_)

    @cached_property
    def _all_sites(self) -> frozenset[BindingSite]:
        """Return all binding sites in the assembly."""
        sites = set()
        for comp_id, component in self._components.items():
            for site in component.site_ids:
                sites.add(BindingSite(component_id=comp_id, site_id=site))
        return frozenset(sites)

    @cached_property
    def _sites_with_bond(self) -> frozenset[BindingSite]:
        """Return a set of binding sites that have bonds."""
        site_with_bonds = set()
        for bond in self.bonds:
            for site in bond.sites:
                site_with_bonds.add(site)
        return frozenset(site_with_bonds)

    @cached_property
    def _component_connection(self) -> frozenset[frozenset[ID]]:
        connections = set()
        for bond in self.bonds:
            comp_ids = frozenset(bond.component_ids)
            connections.add(comp_ids)
        return frozenset(connections)

    def _get_component_of_site(self, site: BindingSite) -> Component:
        """Return the component corresponding to the given binding site."""
        return self._components[site.component_id]

    def _validate(self):
        self._validate_components()
        self._validate_bonds()
        self._validate_parallel_bonds()

    def _validate_components(self):
        # Components with the same kind should have the same structure.
        comp_kind_to_obj: dict[str, Component] = {}
        for comp in self._components.values():
            if comp.kind in comp_kind_to_obj:
                if comp != comp_kind_to_obj[comp.kind]:
                    raise InconsistentComponentError(
                        component_kind=comp.kind,
                        component1=comp,
                        component2=comp_kind_to_obj[comp.kind],
                    )
            else:
                comp_kind_to_obj[comp.kind] = comp

    def _validate_bonds(self):
        component_keys = set(self._components.keys())
        used_sites = set()
        for bond in self.bonds:
            # Validate that the components exist
            for comp_id in bond.component_ids:
                if comp_id not in component_keys:
                    raise InvalidBondError(
                        bond=bond,
                        detail=f"Component {comp_id} not found in assembly.")

            # Validate that the sites exist in the respective components
            for site in bond.sites:
                component = self._components[site.component_id]
                if site.site_id not in component.site_ids:
                    raise InvalidBondError(
                        bond=bond,
                        detail=(
                            f"Site {site.site_id} not found in component "
                            f"{site.component_id}."))

                # Validate that the site is not already used
                if site in used_sites:
                    raise InvalidBondError(
                        bond=bond,
                        detail=f"Site {site} is already used in another bond.")
                used_sites.add(site)

    def _validate_parallel_bonds(self):
        # Currently, parallel bonds (multiple bonds between the same pair
        # of components) are not supported.
        bonded_pairs: dict[frozenset[ID], Bond] = {}
        for bond in self.bonds:
            comp_pair = frozenset(bond.component_ids)
            if comp_pair in bonded_pairs:
                raise ParallelBondError(
                    bond1=bonded_pairs[comp_pair], bond2=bond
                )
            bonded_pairs[comp_pair] = bond



import os
from pathlib import Path

from nasap_net.models import Assembly
from nasap_net.yaml.load import load


def load_from_file(
        file_path: os.PathLike | str,
        *,
        strict: bool = True,
        verbose: bool = True,
) -> list[Assembly]:
    """Load assemblies and components from a YAML file.

    Parameters
    ----------
    file_path : os.PathLike | str
        Path to the YAML file to load.
    strict : bool, optional
        If True, raise an error if the file does not exist.
        If False, return an empty list if the file does not exist.
        Default is True.
    verbose : bool, optional
        If True, print a message indicating the number of assemblies loaded.
        Default is True.

    Returns
    -------
    list[Assembly]
        List of loaded assemblies.
    """
    file_path = Path(file_path)
    if not file_path.exists():
        if strict:
            raise FileNotFoundError(f'File "{str(file_path)}" does not exist.')
        return []

    yaml_str = file_path.read_text(encoding="utf-8")
    assemblies = load(yaml_str)
    if verbose:
        print(f'Loaded {len(assemblies)} assemblies from "{str(file_path)}"')
    return assemblies



import pytest

from nasap_net.models import Assembly, Component
from nasap_net.models.component_consistency_check import \
    InconsistentComponentBetweenAssembliesError, check_component_consistency


def test_valid_case():
    L1 = Component(kind='L', sites=[0, 1])
    L2 = Component(kind='L', sites=[0, 1])
    assemblies = [
        Assembly({'L1': L1}, []),
        Assembly({'L2': L2}, []),
    ]
    # This should not raise an exception
    check_component_consistency(assemblies)


def test_invalid_case():
    L1 = Component(kind='L', sites=[0, 1])
    L2 = Component(kind='L', sites=[2, 3])  # Different
    assemblies = [
        Assembly({'L1': L1}, []),
        Assembly({'L2': L2}, []),
    ]
    with pytest.raises(InconsistentComponentBetweenAssembliesError):
        check_component_consistency(assemblies)



from collections.abc import Iterable, Mapping
from typing import Any

from nasap_net.isomorphism import extract_unique_assemblies_by_isomorphism
from nasap_net.models import Assembly, Component
from nasap_net.types import ID
from .lib import cap_assemblies_with_ligand, enumerate_fragments


def enumerate_assemblies(
        template: Assembly,
        *,
        leaving_ligand: Component,
        leaving_ligand_site: ID | None = None,
        metal_kinds: Iterable[str],
        symmetry_operations: Iterable[Mapping[Any, ID]] | None = None,
) -> set[Assembly]:
    """Enumerate assemblies which can be formed by adding the leaving ligand
    to the fragments of the template assembly.

    This function generates all possible assemblies by attaching the
    specified leaving ligand to each fragment in the template assembly. It
    ensures that the resulting assemblies are unique by excluding duplicates
    based on symmetry operations and isomorphism checks.

    Any two assemblies that meet either of the following conditions are
    considered duplicates, and only one of them is included in the output list:
    - One can be transformed into the other by applying a symmetry operation
      from the provided list.
    - They are isomorphic, i.e., they have the same connectivity structure.

    Providing symmetry operations is strongly recommended to reduce the
    computational cost of duplicate exclusion.

    Parameters
    ----------
    template : Assembly
        The template assembly to base the enumeration on.
    leaving_ligand : Component
        The ligand to be added to the fragments of the template assembly.
    leaving_ligand_site : ID | None, optional
        The site ID on the leaving ligand where it will be attached.
        If None, the first site ID of the leaving ligand will be used.
        Default is None.
    metal_kinds : Iterable[str]
        The kinds of metal components in the assembly. All binding sites on
        these components will be considered for attaching the leaving ligand.
    symmetry_operations : Iterable[Mapping[Any, ID]] | None, optional
        A list of symmetry operations to consider when excluding duplicate
        assemblies. Each symmetry operation is represented as a mapping from
        original site IDs to transformed site IDs.

    Returns
    -------
    set[Assembly]
        A set of unique assemblies formed by adding the leaving ligand
        to the fragments of the template assembly.
        Also includes the free leaving ligand as an assembly.
    """
    fragments = enumerate_fragments(
        template,
        symmetry_operations=symmetry_operations,
    )
    unique_fragments = extract_unique_assemblies_by_isomorphism(
        fragments,
    )
    capped_assemblies = cap_assemblies_with_ligand(
        unique_fragments,
        component=leaving_ligand,
        component_site_id=leaving_ligand_site,
        metal_kinds=metal_kinds,
    )
    # Also include the free leaving ligand as an assembly
    capped_assemblies.add(
        Assembly(
            components={f'{leaving_ligand.kind}0': leaving_ligand},
            bonds=[],
        )
    )
    return capped_assemblies



from collections import defaultdict
from collections.abc import Hashable, Iterable

from nasap_net.models import Assembly
from .is_isomorphic import is_isomorphic
from .lib import get_light_signature


def extract_unique_assemblies_by_isomorphism(
        assemblies: Iterable[Assembly]
) -> set[Assembly]:
    """Extract unique assemblies by isomorphism from a collection of assemblies.

    Parameters
    ----------
    assemblies : Iterable[Assembly]
        The collection of assemblies to extract unique assemblies from.

    Returns
    -------
    set[Assembly]
        A set of unique assemblies by isomorphism.
    """
    sig_to_unique_assembly: dict[Hashable, set[Assembly]] = defaultdict(set)
    for assembly in assemblies:
        sig = get_light_signature(assembly)
        is_unique = True
        for unique_assembly in sig_to_unique_assembly[sig]:
            if is_isomorphic(assembly, unique_assembly):
                is_unique = False
                break
        if is_unique:
            sig_to_unique_assembly[sig].add(assembly)
    unique_assemblies = set()
    for assemblies_set in sig_to_unique_assembly.values():
        unique_assemblies.update(assemblies_set)
    return unique_assemblies



from nasap_net.models import Assembly
from .exceptions import NoIsomorphismFoundError
from .lib import color_vertices_and_edges, convert_assembly_to_igraph, \
    decode_mapping
from .models import Isomorphism


def get_isomorphism(assem1: Assembly, assem2: Assembly) -> Isomorphism:
    conv_res1 = convert_assembly_to_igraph(assem1)
    conv_res2 = convert_assembly_to_igraph(assem2)

    g1 = conv_res1.graph
    g2 = conv_res2.graph

    try:
        colors = color_vertices_and_edges(g1, g2)
    except NoIsomorphismFoundError:
        raise NoIsomorphismFoundError() from None

    mapping: list[int]
    _, mapping, _ = g1.isomorphic_vf2(
        g2,
        color1=colors.v_color1,
        color2=colors.v_color2,
        edge_color1=colors.e_color1,
        edge_color2=colors.e_color2,
        return_mapping_12=True,
    )

    return decode_mapping(mapping, conv_res1, conv_res2)


def get_all_isomorphisms(
        assem1: Assembly, assem2: Assembly) -> set[Isomorphism]:
    conv_res1 = convert_assembly_to_igraph(assem1)
    conv_res2 = convert_assembly_to_igraph(assem2)

    try:
        colors = color_vertices_and_edges(conv_res1.graph, conv_res2.graph)
    except NoIsomorphismFoundError:
        raise NoIsomorphismFoundError() from None

    res: list[list[int]] = conv_res1.graph.get_isomorphisms_vf2(
        conv_res2.graph,
        color1=colors.v_color1,
        color2=colors.v_color2,
        edge_color1=colors.e_color1,
        edge_color2=colors.e_color2,
    )

    return {decode_mapping(mapping, conv_res1, conv_res2) for mapping in res}



from collections.abc import Iterable
from dataclasses import dataclass
from typing import TypeVar

from nasap_net.models import Assembly
from nasap_net.types import ID
from nasap_net.yaml.lib import dump_components, dump_semi_light_assemblies
from nasap_net.yaml.semi_light_assembly import \
    convert_assemblies_to_semi_light_ones

_T = TypeVar('_T', bound=ID)

def dump(assemblies: Iterable[Assembly]) -> str:
    """Dump assemblies and components into a YAML string."""
    dumped = _dump_separately(assemblies)
    return '---\n'.join([dumped.components, dumped.assemblies])


@dataclass(frozen=True)
class _Dumped:
    assemblies: str
    components: str


def _dump_separately(
        assemblies: Iterable[Assembly],
        ) -> _Dumped:
    assembly_list = list(assemblies)
    assembly_map = dict(enumerate(assembly_list))
    res = convert_assemblies_to_semi_light_ones(assembly_map)

    res_assemblies = [
        res.semi_light_assemblies[id_]
        for id_ in sorted(res.semi_light_assemblies.keys())
    ]

    return _Dumped(
        assemblies=dump_semi_light_assemblies(
            res_assemblies
        ),
        components=dump_components(
            dict(sorted(res.components.items())),
        ),
    )



from nasap_net.models import Assembly, Component
from nasap_net.yaml.helper import split_yaml_documents
from nasap_net.yaml.lib import load_components, load_semi_light_assemblies
from .semi_light_assembly import SemiLightAssembly, \
    convert_semi_light_assemblies_to_rich_ones


def load(yaml_str: str) -> list[Assembly]:
    components, light_assemblies = _load_components_and_semi_light_assemblies(
        yaml_str=yaml_str,
    )
    assembly_map = convert_semi_light_assemblies_to_rich_ones(
        dict(enumerate(light_assemblies)), components
    )
    return list(assembly_map.values())


def _load_components_and_semi_light_assemblies(
        yaml_str: str,
) -> tuple[dict[str, Component], list[SemiLightAssembly]]:
    docs = list(split_yaml_documents(yaml_str))
    if len(docs) != 2:
        raise ValueError(
            f"Expected exactly 2 YAML documents, found {len(docs)}.")
    components = load_components(docs[0])
    semi_light_assemblies = load_semi_light_assemblies(docs[1])
    return components, semi_light_assemblies



from __future__ import annotations

from dataclasses import dataclass

from nasap_net.types import ID
from .light_bond import LightBond


@dataclass(frozen=True)
class GrowingStep:
    bond_to_add: LightBond
    component_to_add: ID | None



import pytest

from nasap_net.models import Assembly, Bond, Component
from nasap_net.models.assembly import InconsistentComponentError


def test_assembly():
    L = Component(kind='L', sites=[0, 1])
    M = Component(kind='M', sites=[0, 1])
    assembly = Assembly(
        components={'L1': L, 'M1': M},
        bonds={Bond('L1', 0, 'M1', 0)},
        )
    assert assembly.components == {'L1': L, 'M1': M}
    assert assembly.bonds == frozenset({Bond('L1', 0, 'M1', 0)})


def test_inconsistent_component_kind_error():
    L1 = Component(kind='L', sites=[0, 1])
    L2 = Component(kind='L', sites=[2, 3])
    with pytest.raises(InconsistentComponentError):
        Assembly(
            components={'L0': L1, 'L1': L2},
            bonds=set()
        )


def test___repr__():
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    assembly = Assembly(
        components={'X0': X, 'M0': M, 'X1': X},
        bonds={Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)},
    )
    expected_repr = (
        "<Assembly components={'M0': 'M', 'X0': 'X', 'X1': 'X'}, "
        "bonds=[('M0', 0, 'X0', 0), ('M0', 1, 'X1', 0)]>"
    )
    assert repr(assembly) == expected_repr


def test___repr___with_id():
    X = Component(kind='X', sites=[0])
    assembly = Assembly(
        id_='free_X',
        components={'X0': X},
        bonds=[],
    )
    expected_repr = (
        "<Assembly id='free_X', components={'X0': 'X'}, bonds=[]>"
    )
    assert repr(assembly) == expected_repr


def test___repr___with_no_bonds():
    X = Component(kind='X', sites=[0])
    assembly = Assembly(
        components={'X0': X},
        bonds=[],
    )
    expected_repr = (
        "<Assembly components={'X0': 'X'}, bonds=[]>"
    )
    assert repr(assembly) == expected_repr
